import pytest
from wscodec.encoder.pyencoder.instrumented import InstrumentedSampleTRH
from wscodec.decoder import decode

INPUT_SERIAL = 'abcdabcd'
INPUT_TIMEINT = 12
INPUT_SECKEY = 'AAAABBBBCCCCDDDD'
INPUT_STATUSB64 = 'MDAwMDAw'
INPUT_CBUFLENBLKS = 32

PADDING = 2


@pytest.fixture(scope="function",
                params=[{'baseurl': "thehrhsdbsdrd.plotsensor.com", 'secretkey': INPUT_SECKEY},
                        {'baseurl': "plotsensor.com", 'secretkey': INPUT_SECKEY}
                        ])
def instr_sample(request):
    return InstrumentedSampleTRH(baseurl=request.param['baseurl'],
                              serial=INPUT_SERIAL,
                              secretkey=request.param['secretkey'],
                              smplintervalmins=INPUT_TIMEINT)


@pytest.mark.parametrize('n', range(1, 150))
def test_cursorpos(instr_sample, n):
    instr_sample.pushsamples(n)
    encoder_cursorpos = instr_sample.ffimodule.lib.demi_getendmarkerpos()

    # Decode the URL
    par = instr_sample.eepromba.get_url_parsedqs()
    decodedurl = decode(secretkey=instr_sample.secretkey, statb64=par['x'][0], timeintb64=par['t'][0],
                        circb64=par['q'][0], vfmtb64=par['v'][0])

    assert encoder_cursorpos == decodedurl.endmarkerpos



#  Copyright (c) Plotsensor Ltd. 2020.
from .circularbuffer import CircularBufferURL
from .exceptions import MessageIntegrityError
from .b64decode import B64Decoder
from typing import List
from enum import Enum
import hashlib
import hmac

BYTES_PER_PAIR = 3                                      #: The number of bytes in each decoded Pair.
BYTES_PER_PAIRB64 = 4                                   #: The number of bytes in each base64 encoded Pair.
PAIRS_PER_DEMI = 2                                      #: The number of pairs in each 8-byte demi.
BYTES_PER_DEMI = BYTES_PER_PAIRB64 * PAIRS_PER_DEMI     #: The number of bytes in each demi.


class HashType(Enum):
    MD5 = 1
    HMAC_MD5 = 2


class Pair:
    """
    Class representing a pair of 12-bit sensor readings.

    In the URL each pair consists of a 4 byte (base64) string. These decode to 3 8-bit bytes. The last contains
    4-bits from reading0 and 4-bits from reading 1.

    When this class is instantiated, the 3 8-bit bytes are converted back into two 12-bit readings.

    Parameters
    ----------
    rd0MSB : int
        Most Signficant Byte of environmental sensor reading0.
    rd1MSB : int
        Most Significant Byte of environmental sensor reading1.
    Lsb : int
        Upper 4-bits are the least significant bits of reading0.
        Lower 4-bits are the least significant bits of reading1.
    """
    def __init__(self, rd0MSB: int, rd1MSB: int, Lsb: int):

        self.rd0MSB = rd0MSB
        self.rd1MSB = rd1MSB
        self.Lsb = Lsb

        rd0Lsb = (self.Lsb >> 4) & 0xF
        rd1Lsb = self.Lsb & 0xF
        self.rd0 = ((self.rd0MSB << 4) | rd0Lsb)
        self.rd1 = ((self.rd1MSB << 4) | rd1Lsb)

    def __repr__(self):
        """

        Returns
        -------
        A string containing both 12-bit readings.
        """
        return str(self.readings())

    @classmethod
    def from_bytes(cls, bytes: bytes):
        """

        Parameters
        ----------
        bytes : bytes
            The 3 bytes that make up a pair rd0MSB, rd1MSB and Lsb.

        Returns
        -------
        A pair instantiated from the 3 byte input.

        """
        assert len(bytes) == BYTES_PER_PAIR
        return cls(rd0MSB=bytes[0], rd1MSB=bytes[1], Lsb=bytes[2])

    @classmethod
    def from_b64(cls, pairb64: str):
        """

        Parameters
        ----------
        str4: str
            A 4 character string that represents a base64 encoded pair. These are extracted from the circular buffer.

        Returns
        -------
        A pair instantiated from the 4 character string.
        """
        assert len(pairb64) == BYTES_PER_PAIRB64
        pairbytes = B64Decoder.b64decode(pairb64)
        return cls.from_bytes(pairbytes)

    def readings(self):
        """

        Returns
        --------
        A dictionary containing both 12-bit readings.
        """
        return {'rd0': self.rd0, 'rd1': self.rd1}


class PairsURL(CircularBufferURL):
    """
    This takes the payload of the linearised buffer, which is a long string of base64 characters. It decodes this
    into a list of pairs. The hash (MD5 or HMAC-MD5) is taken and compared with that supplied in the URL by the
    encoder. If the hashes match then the decode has been successful. If not, an exception is raised.

    Parameters
    ----------
    *args
        Variable length argument list.
    usehmac: bool
        True if the hash inside the circular buffer endstop is HMAC-MD5. False if it is MD5.
    secretkey: str
        HMAC secret key as a string. Normally 16 characters long.
    **kwargs
        Keyword arguments to be passed to parent class constructors.
    """
    def __init__(self, *args, usehmac: bool = False, secretkey: str = None,  **kwargs):
        super().__init__(*args, **kwargs)

        self._decode_pairs()
        self._verify(usehmac, secretkey)

    def _verify(self, usehmac : bool, secretkey : str):
        """
        Calculate a hash from the list of pairs according to the same algorithm used
        by the encoder (see :ref:`pairhist_hash`). Besides pairs, data from the status URL parameter
        are included. This makes it very unlikely that the same data will be hashed twice, as well as 'protecting'
        the status parameter from modification by a 3rd party.

        A fragment of the calculated hash is compared with that supplied by the encoder. If the hashes agree then
        verification is successful. If not, an exception is raised.

        Parameters
        ----------
        usehmac : bool
            True if the hash inside the circular buffer endstop is HMAC-MD5. False if it is MD5.
        secretkey : str
            HMAC secret key as a string. Normally 16 characters long.

        Raises
        -------
            MessageIntegrityError: If the hash calculated by this decoder does not match the hash provided by the encoder.

        """
        pairhist = bytearray()

        for pair in self.pairs:
            pairhist.append(pair.rd0MSB)
            pairhist.append(pair.rd1MSB)
            pairhist.append(pair.Lsb)

        pairhist.append(self.status.loopcount >> 8)
        pairhist.append(self.status.loopcount & 0xFF)
        pairhist.append(self.status.resetsalltime >> 8)
        pairhist.append(self.status.resetsalltime & 0xFF)
        pairhist.append(self.status.batv_resetcause >> 8)
        pairhist.append(self.status.batv_resetcause & 0xFF)
        pairhist.append(self.endmarkerpos >> 8)
        pairhist.append(self.endmarkerpos & 0xFF)

        # Perform message authentication.
        calcHash, self.hashtype = self.__class__._gethash(pairhist, usehmac, secretkey)
        urlHash = self.hash

        # Truncate calculated MD5 to the same length as the URL MD5.
        calcHash = calcHash[0:len(urlHash)]

        if urlHash != calcHash:
            raise MessageIntegrityError(calcHash, urlHash)

        assert self.npairs == len(self.pairs)

    @staticmethod
    def _gethash(message: bytearray, usehmac: bool, secretkey: str):
        """
        Calculates the hash of a message.

        Parameters
        ----------
        message : bytearray
            Input data to the hashing algorithm.
        usehmac : bool
            When True the HMAC-MD5 algorithm is used. Otherwise MD5 is used (not recommended for production).
        secretkey : str
            HMAC secret key as a string. Normally 16 characters long.

        Returns
        -------
        digest : str
            The message hash.
        hashtype : HashType
            The hash algorithm used.

        """
        secretkeyba = bytearray(secretkey, 'utf8')
        if usehmac:
            hmacobj = hmac.new(secretkeyba, message, "md5")
            digest = hmacobj.hexdigest()
            hashtype = HashType.HMAC_MD5
        else:
            digest = hashlib.md5(message).hexdigest()
            hashtype = HashType.MD5
        return digest, hashtype

    def _decode_pairs(self):
        """
        The payload string is converted into a list of 8-byte demis (see :ref:`demi`).

        The first demi is the newest; its data have been written to the circular buffer most recently,
        so it closest to the left of the endstop. It can contain either one or two pairs.
        This is decoded first.

        Subsequent (older) demis each contain 2 pairs. These are decoded. The final list of pairs is in
        chronological order with the newest first and the oldest last.
        """
        self.pairs = list()

        # Convert payload string into 8 byte demis.
        demis = self._dividestring(self.payloadstr, BYTES_PER_DEMI)

        # The newest 8 byte demi might only contain 1 valid pair.
        # If so, it is a partial one so it gets processed first.
        partial = self.npairs % PAIRS_PER_DEMI
        full = int(self.npairs / PAIRS_PER_DEMI)

        if partial != 0:
            demi = demis.pop()
            pair = self._pairsfromdemi(demi)[1]  # Only append the oldest pair, for the newest is invalid.
            self.pairs.append(pair)

        # Process remaining full demis. These all contain 2 pairs.
        for i in range(0, full, 1):
            demi = demis.pop()
            demipairs = self._pairsfromdemi(demi)  # Append both pairs.
            self.pairs.extend(demipairs)

    @staticmethod
    def _dividestring(source: str, n: int):
        """

        Parameters
        ----------
        source : str
            The string to be divided.
        n
            The number of characters in each substring.

        Returns
        -------
        A list of substrings, each containing n characters.

        """
        return list(source[i:i+n] for i in range(0, len(source), n))

    def _pairsfromdemi(self, demi: str) -> List[Pair]:
        """
        Decode a demi into 2 pairs.

        Parameters
        ----------
        demi : str
            A string containing 8 base64 characters.

        Returns
        -------
        A list of 2 pairs:
            Element 0 is the oldest pair, decoded from the first 4 demi characters.
            Element 1 is the newest pair, decoded from the last 4 demi characters.

        """
        pairs = list()

        assert len(demi) == BYTES_PER_DEMI, demi

        for i in range(0, BYTES_PER_DEMI, BYTES_PER_PAIRB64):
            pairb64 = demi[i:i+BYTES_PER_PAIRB64]
            pair = Pair.from_b64(pairb64)
            pairs.append(pair)

        pairs.reverse()
        return pairs


import pytest
from wscodec.decoder import decode


@pytest.fixture(scope="function", params=[1, 10, 100, 1000, 1001])
def instr_sample_populated(instr_sample, request):
    inlist = instr_sample.pushsamples(request.param)

    # Decode the URL
    par = instr_sample.eepromba.get_url_parsedqs()
    decodedurl = decode(secretkey=instr_sample.secretkey, statb64=par['x'][0], timeintb64=par['t'][0],
                        circb64=par['q'][0], vfmtb64=par['v'][0])

    urllist = decodedurl.get_samples_list()
    for d in urllist:
        del d['timestamp']
        del d['rawtemp']
        if 'rawrh' in d:
            del d['rawrh']

    inlist = inlist[:len(urllist)]

    return {'instr_sample': instr_sample, 'inlist': inlist, 'urllist': urllist}


import pytest
from wscodec.encoder.pyencoder.instrumented import InstrumentedSampleTRH
from wscodec.decoder import decode

INPUT_SERIAL = 'abcdabcd'
INPUT_TIMEINT = 65535 # Maximum time interval
INPUT_SECKEY = 'AAAABBBBCCCCDDDD'
INPUT_STATUSB64 = 'MDAwMDAw'
INPUT_BASEURL = 'plotsensor.com'
INPUT_CBUFLENBLKS = 32

PADDING = 2


@pytest.fixture(scope="function",
                params=[{'baseurl': "plotsensor.com", 'secretkey': INPUT_SECKEY}
                        ])
def instr_sample(request):
    return InstrumentedSampleTRH(baseurl=request.param['baseurl'],
                              serial=INPUT_SERIAL,
                              secretkey=request.param['secretkey'],
                              smplintervalmins=INPUT_TIMEINT)


def test_elapsedmins(instr_sample):
    instr_sample.pushsamples(1)

    for i in range(0, INPUT_TIMEINT, 5000):
        instr_sample.ffimodule.lib.enc_setelapsed(i)

        # Decode the URL
        par = instr_sample.eepromba.get_url_parsedqs()
        decodedurl = decode(secretkey=instr_sample.secretkey, statb64=par['x'][0], timeintb64=par['t'][0],
                            circb64=par['q'][0], vfmtb64=par['v'][0])

        assert i == decodedurl.elapsedmins






# Helpful: https://julien.danjou.info/python-exceptions-guide/

class DecoderError(Exception):
    """Base cupl Decoder error class."""

    def __init__(self, msg="cupl Decoder Error"):
        super().__init__(msg)


class InvalidMajorVersionError(DecoderError):
    def __init__(self, encoderversion, decoderversion, msg=None):
        if msg is None:
            # Set a default error messasge
            msg = "There is a mismatch between the encoder version = {} " \
                  "and decoder version = {} ".format(encoderversion, decoderversion)
        super().__init__(msg)
        self.encoderversion = encoderversion
        self.decoderversion = decoderversion


class InvalidFormatError(DecoderError):
    def __init__(self, circformat, msg=None):
        if msg is None:
            # Set a default error message
            msg = "Invalid circular buffer format = {}.".format(circformat)
        super().__init__(msg)
        self.circformat = circformat


class MessageIntegrityError(DecoderError):
    def __init__(self, calchash, urlhash, msg=None):
        if msg is None:
            # Set a default error message
            msg = "Checksum mismatch. Calculated hash = {}, URL hash = {}".format(calchash, urlhash)
        super().__init__(msg)
        self.calchash = calchash
        self.urlhash = urlhash


class NoCircularBufferError(DecoderError):
    def __init__(self, status, msg=None):
        if msg is None:
            # Set a default error message
            msg = "There is no circular buffer. This is indicative of an error with the system running the encoder. " \
                  "Status = {}".format(status)
        super().__init__(msg)
        self.status = status


class DelimiterNotFoundError(DecoderError):
    def __init__(self, circb64, status, msg=None):
        if msg is None:
            msg = " No delimiting character found in the circular buffer string = {}. The firmware has initialised " \
                  "the encoder but not pushed data. Status = {}".format(circb64, status)
        super().__init__(msg)
        self.circb64 = circb64
        self.status = status



import ndef
import urllib.parse as urlparse
from ctypes import c_char_p


class Eeprom():
    """
    A mock of the NT3H2111 EEPROM, based on a bytearray. There are methods to read and write from this in 16-byte blocks.
    Helper methods parse the entire EEPROM contents as an NDEF message. This mimics what a phone will do when it reads
    the NT3H2111 using NFC.

    The EEPROM will normally contain the output of the cupl Encoder: an NDEF message containing one NDEF record.
    This itself will contain a URL. One of the parameters in the query string is 'q', which contains the circular
    buffer of temperature and relative humidity samples.
    """
    def __init__(self, sizeblocks: int):
        """
        Instatiate a mock EEPROM.

        :param sizeblocks: Size of the EEPROM in 16-byte blocks.
        """
        self.eepromba = bytearray(sizeblocks*16)
        self.sizeblocks = sizeblocks

    def write_block(self, block: int, blkdata: c_char_p):
        """
        Write one block from a pointer.

        :param block: Address of the block to write
        :param blkdata: Pointer to an array of 16 bytes that will be written to the block.
        :return: None
        """
        blkstartbyte = block*16
        for i in range(0,16):
            self.eepromba[blkstartbyte + i] = int.from_bytes(blkdata[i], byteorder="little")

    def read_block(self, block: int, blkdata: c_char_p):
        """
        Read one block into a pointer.

        :param block: Address if the block to read.
        :param blkdata: Pointer to an array of 16 bytes that the block will be read into.
        :return: None
        """
        blkstartbyte = block*16
        for i in range(0,16):
            blkdata[i] = bytes(self.eepromba[blkstartbyte+i:blkstartbyte+i+1])
        return 0

    def display_block(self, block: int) -> bytearray:
        """
        Display one EEPROM block.

        :param block: Address of the block to display.
        :return: Block data as a list of 16 bytes.
        """
        blkstartbyte = block*16
        blkendbyte = blkstartbyte + 16
        return self.eepromba[blkstartbyte:blkendbyte]

    def __repr__(self) -> str:
        """
        Display all EEPROM blocks.

        :return: The contents of all EEPROM blocks as a string.
        """
        retstr = "EEPROM with " + str(self.sizeblocks) + " blocks\n"
        for block in range(0, self.sizeblocks):
            retstr += "EEPROM block " + str(block) + "\n"
            blkstartbyte = block*16
            blkendbyte = blkstartbyte + 16
            retstr += str(self.eepromba[blkstartbyte:blkendbyte]) + "\n"
        return retstr

    def get_message(self) -> bytearray:
        """
        Extract the NDEF message bytes from EEPROM.

        :return: NDEF message bytearray
        """
        ndefmsgstartbyte = 16 + 4
        return self.eepromba[ndefmsgstartbyte:]

    def decode_ndef(self) -> ndef.Record:
        """
        Decode the NDEF message.

        :return: First NDEF record in the message
        """
        sizedemis = self.sizeblocks * 2
        ndefmessage = self.get_message()
        # Copy NDEF message into a bytearray
        ndefdemis = bytearray()
        for i in range(0,sizedemis):
            startbyte = i*8
            endbyte = startbyte + 8
            demi = ndefmessage[startbyte:endbyte]
            ndefdemis += demi
        # Decode the bytearray.
        ndefrecords = ndef.message_decoder(ndefdemis)
        # Use next to generate the first record from the ndefrecords generator
        return next(ndefrecords)

    def get_url(self) -> str:
        """
        Obtain URL from the NDEF record stored in EEPROM.

        :return: URL string
        """
        ndefmessage = self.decode_ndef()
        return ndefmessage.iri

    def get_url_parsed(self) -> urlparse.ParseResult:
        """
        Parse URL in the EEPROM NDEF record.

        :return: Parsed URL
        """
        url = self.get_url()
        return urlparse.urlparse(url)

    def get_url_parsedqs(self) -> dict:
        """
        Parse parameters in the `query string <https://en.wikipedia.org/wiki/Query_string>`_

        :return: A dictionary of URL parameters.
        """
        parsedurl = self.get_url_parsed()
        urlquery = parsedurl.query
        return urlparse.parse_qs(urlquery)

    def get_qparam(self) -> str:
        """
        :return: The value of URL parameter 'q' as a string.
        """
        parsedqs = self.get_url_parsedqs()
        q = parsedqs['q']
        return q[0]

    def get_qdemis(self) -> list:
        """
        :return: The value of URL parameter 'q' as a list of 8-byte demis.
        """
        qstr = self.get_qparam()
        n = 8
        qdemis = [qstr[i:i+n] for i in range(0, len(qstr), n)]
        return qdemis



from datetime import datetime
from .samples import SamplesURL, Sample


class TempSample(Sample):
    def __init__(self, rawtemp: int, timestamp: datetime):
        """

        :param rawtemp:
        :param timestamp:
        """
        super().__init__(timestamp)
        self.rawtemp = rawtemp
        self.temp = self.reading_to_temp(rawtemp)

    @staticmethod
    def reading_to_temp(reading: int) -> float:
        """
        :param reading: Integer temperature ADC reading from the HDC2021.
        :return: Temperature in degrees C
        """
        return (reading * 165) / 4096 - 40


class TempRHSample(TempSample):
    def __init__(self, rawtemp: int, rawrh: int, timestamp: datetime):
        super().__init__(rawtemp, timestamp)
        self.rawrh = rawrh
        self.rh = self.reading_to_rh(rawrh)

    @staticmethod
    def reading_to_rh(reading: int) -> float:
        """

        :param reading: Integer Relative Humidity ADC reading from the HDC2021.
        :return: Relative Humidity in percent.
        """
        return (reading * 100) / 4096


class TempRH_URL(SamplesURL):
    def __init__(self, *args, **kwargs):
        """
        :return: Decoded URL parameters with buffer data converted to a list of timestamped samples, each containing
        temperature (degrees C) and relative humidity (%) readings.
        """
        super().__init__(*args, **kwargs)
        timestamp_gen = self.generate_timestamp()

        for pair in self.pairs:
            temp = pair.rd0
            rh = pair.rd1

            sample = TempRHSample(temp, rh, timestamp=next(timestamp_gen))
            self.samples.append(sample)


class Temp_URL(SamplesURL):
    def __init__(self, *args, **kwargs):
        """

        :return: Decoded URL parameters with buffer data converted to a list of timestamped samples, each containing one
        temperature reading in degrees C.
        """
        super().__init__(*args, **kwargs)
        timestamp_gen = self.generate_timestamp()

        for pair in self.pairs:
            if pair.rd1 != 4095:
                sample = TempSample(pair.rd1, timestamp=next(timestamp_gen))
                self.samples.append(sample)

            sample = TempSample(pair.rd0, timestamp=next(timestamp_gen))
            self.samples.append(sample)



from datetime import datetime
import pkg_resources
from .samples import SamplesURL
from .b64decode import B64Decoder
from .exceptions import InvalidMajorVersionError, InvalidFormatError
from . import hdc2021


def decode(secretkey: str,
           statb64: str,
           timeintb64: str,
           circb64: str,
           vfmtb64: str,
           usehmac: bool = True,
           scantimestamp: datetime = None) -> SamplesURL:
    """
    Decode the version string and extract codec version and format code. An error is raised if the codec version does
    not match. A decoder object is returned based on the format code. An error is raised if no decoder is available
    for the code.

    Parameters
    -----------
    secretkey: str
        HMAC secret key as a string. Normally 16 characters long.

    statb64: str
        Value of the URL parameter that holds status information (base64 encoded).

    timeintb64: str
        Value of the URL parameter that holds the time interval between samples in minutes (base64 encoded).

    circb64: str
        Value of the URL parameter that contains the circular buffer of base64 encoded samples.

    vfmtb64: str
        Value of the URL parameter that contains the version and format string (base64 encoded).

    usehmac: bool
        True if the hash inside the circular buffer endstop is HMAC-MD5. False if it is MD5.

    scantimestamp: datetime
        The time that the tag was scanned. All decoded samples will be timestamped relative to this.

    Returns
    --------
    SamplesURL
        An object containing a list of timestamped environmental sensor samples.

    """

    encodermajorversion, formatcode = _get_encoderversion(vfmtb64)
    decodermajorversion = _get_decoderversion()

    if encodermajorversion != decodermajorversion:
        raise InvalidMajorVersionError(encodermajorversion, decodermajorversion)

    decoder = _get_decoder(formatcode)(statb64=statb64, timeintb64=timeintb64, circb64=circb64, usehmac=usehmac, secretkey=secretkey, scantimestamp=scantimestamp)
    return decoder


def _get_encoderversion(vfmtb64):
    vfmtb64 = vfmtb64[-4:]
    vfmtbytes = B64Decoder.b64decode(vfmtb64)
    encoderbytes = vfmtbytes[0:2]
    encoderversion = int.from_bytes(encoderbytes, byteorder='big')
    formatcode = vfmtbytes[2]
    return encoderversion, formatcode


def _get_decoderversion():
    decoderversionstr = pkg_resources.require("cuplcodec")[0].version
    decodermajorversion = int(decoderversionstr.split('.', 1)[0])
    return decodermajorversion


def _get_decoder(formatcode: int):
    """
        Parameters
        -----------
        formatcode:
            Value of the codec format field. Specifies which decoder shall be returned.

        Return
        -------
        Decoder class for the given format code.

        """
    decoders = {
        1: hdc2021.TempRH_URL,
        2: hdc2021.Temp_URL
    }
    try:
        decoder = decoders[formatcode]
    except KeyError:
        raise InvalidFormatError(formatcode)

    return decoder








from .b64decode import B64Decoder
import struct

BIT0 = 0x01
BIT1 = 0x02
BIT2 = 0x04
BIT3 = 0x08
BIT4 = 0x10
BIT5 = 0x20
BIT6 = 0x40
BIT7 = 0x80

BOR_BIT = BIT0
SVSH_BIT = BIT1
WDT_BIT = BIT2
MISC_BIT = BIT3
LPM5WU_BIT = BIT4
CLOCKFAIL_BIT = BIT5
SCANTIMEOUT_BIT = BIT7


class Status:
    """
        Decode the status string.

        Parameters
        -----------
        statb64:
            Value of the URL parameter that holds status information (after base64 encoding).

    """
    def __init__(self, statb64: str):
        decstr = B64Decoder.b64decode(statb64)
        declist = struct.unpack("HHH", decstr)
        self.loopcount = declist[0]
        self.resetsalltime = declist[1]
        self.batv_resetcause = declist[2]

        brownout = (self.get_resetcauseraw() & BOR_BIT) > 0
        supervisor = (self.get_resetcauseraw() & SVSH_BIT) > 0
        watchdog = (self.get_resetcauseraw() & WDT_BIT) > 0
        misc = (self.get_resetcauseraw() & MISC_BIT) > 0
        lpm5wakeup = (self.get_resetcauseraw() & LPM5WU_BIT) > 0
        clockfail = (self.get_resetcauseraw() & CLOCKFAIL_BIT) > 0
        scantimeout = (self.get_resetcauseraw() & SCANTIMEOUT_BIT) > 0

        self.resetcause = {
            "brownout": brownout,
            "supervisor": supervisor,
            "watchdog": watchdog,
            "misc": misc,
            "lpm5wakeup": lpm5wakeup,
            "clockfail": clockfail,
            "scantimeout": scantimeout
        }

    def __str__(self):
        return "Reset cause: {}, " \
               "Loop count: {}, " \
               "Resets all time: {}, " \
               "Battery voltage: {}".format(self.resetcause, self.loopcount, self.resetsalltime, self.get_batvoltagemv())

    def get_batvoltageraw(self):
        """

        :return: Battery voltage as an 8-bit value.
        """
        return self.batv_resetcause >> 8

    def get_resetcauseraw(self):
        """

        :return: Reset cause as an 8-bit value.
        """
        return self.batv_resetcause & 0xFF

    def get_batvoltagemv(self):
        """

        :return: Battery voltage converted to mV.
        """
        return (256 * 1500) / self.get_batvoltageraw()





import cffi
import subprocess
import re
import pycparser
from pycparser import c_generator
import os
import importlib
import weakref
import pkg_resources

resource_package = __package__
print(__package__)
c_encoder_path = '/'.join(('', 'c_encoder', ''))
pycparser_path = '/'.join(('', 'pycparser', ''))
sharedobj_path = '/'.join(('', 'sharedobj', ''))

ENCODER_CSOURCE_PATH = pkg_resources.resource_filename(resource_package, c_encoder_path)
PYCPARSER_PATH = pkg_resources.resource_filename(resource_package, pycparser_path)
SHAREDOBJ_PATH = pkg_resources.resource_filename(resource_package, sharedobj_path)

global_weakkeydict = weakref.WeakKeyDictionary()


class CFFIGenerator(pycparser.c_generator.CGenerator):
    def __init__(self, blacklist):
        super().__init__()
        self.blacklist = blacklist

    def visit_Decl(self, n, *args, **kwargs):
        result = super().visit_Decl(n, *args, **kwargs)
        if isinstance(n.type, pycparser.c_ast.FuncDecl):
            if n.name not in self.blacklist:
                return 'extern "Python+C" ' + result
        return result


def convert_function_declarations(source, blacklist):
    return CFFIGenerator(blacklist).visit(pycparser.CParser().parse(source))


class FunctionList(pycparser.c_ast.NodeVisitor):
    def __init__(self, source):
        self.funcs = set()
        self.visit(pycparser.CParser().parse(source))

    def update(self, otherlist):
        self.funcs.update(otherlist)

    def visit_FuncDef(self, node):
        self.funcs.add(node.decl.name)


def preprocess(source):
    return subprocess.run(['gcc', '-E', '-P', '-', '-I'+ENCODER_CSOURCE_PATH, '-I'+PYCPARSER_PATH+'/utils/fake_libc_include'],
                          input=source, stdout=subprocess.PIPE,
                          universal_newlines=True, check=True).stdout


def load(filename, depfilenames=list()):
 """
 Load a file
 """
 name = __package__ + '.sharedobj.' + filename + 'py'
 # load source code
 source = open(ENCODER_CSOURCE_PATH + filename + '.c').read()

 for depfilename in depfilenames:
     depsource = open(ENCODER_CSOURCE_PATH + depfilename + '.c').read()
     source = source + depsource

 # preprocess all header files for CFFI
 includes = preprocess(''.join(re.findall('\s*#include\s+.*', source)))

 # Obtain a list of local functions
 local_functions = FunctionList(preprocess(source)).funcs
 # prefix external functions (not in the list) with extern "Python+C"
 includes = convert_function_declarations(includes, local_functions)

 # Obtain the absolute directory path of this file.
 abspath = os.path.dirname(__file__)

 # Pass source code and libraries to FFI Builder
 ffibuilder = cffi.FFI()

 # Add all includes to the cdef attribute
 ffibuilder.cdef(includes + "nv_t nv;", packed=False)

 print("SOURCE " + source)

 # Add the source, library sources and include directories
 ffibuilder.set_source(name, source, include_dirs=[ENCODER_CSOURCE_PATH])

 return ffibuilder



from .pairs import PairsURL
from .b64decode import B64Decoder
from datetime import timedelta, timezone, datetime


class Sample:
    """
    Sample base class. All samples must contain a timestamp.
    """
    def __init__(self, timestamp: datetime):
        """

        Parameters
        ----------
        timestamp : datetime
            Estimated sample collection time. Accurate to one minute.
        """
        self.timestamp = timestamp


class SamplesURL(PairsURL):
    """
    This holds a list of decoded sensor samples. Each needs a timestamp, but this must be calculated. There
    are no absolute timestamps in the URL. First, all times are relative to scantimestamp (when the tag was
    scanned).

    The URL does contain self.elapsedmins (the minutes elapsed since the newest sample was acquired). This makes it
    possible to calculate self.newest_timestamp, the timestamp of the newest sample.

    Every subsequent sample is taken at a fixed time interval relative to self.newest_timestamp. This time interval
    is decoded from timeintb64 into self.timeinterval.

    Parameters
    ----------
    *args
        Variable length argument list
    timeintb64 : str
        Time interval between samples in minutes, base64 encoded into a 4 character string.
    scantimestamp : datetime
        Time the tag was scanned. It corresponds to the time the URL on the tag is requested from the web server.
    **kwargs
        Keyword arguments to be passed to parent class constructors.
    """
    def __init__(self, *args, timeintb64: str, scantimestamp: datetime = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.timeintb64 = timeintb64
        self.scantimestamp = scantimestamp or datetime.now(timezone.utc)

        # Calculates the time interval in minutes from a URL parameter.
        timeintmins_bytes = B64Decoder.b64decode(self.timeintb64)
        self.timeintmins_int = int.from_bytes(timeintmins_bytes, byteorder='little')

        # Convert time interval into a timedelta object.
        self.timeinterval = timedelta(minutes=self.timeintmins_int)
        # Calculate the timestamp of the newest sample
        self.newest_timestamp = self.scantimestamp - timedelta(minutes=self.elapsedmins)
        # Define an empty list to hold samples.
        self.samples = list()

    def get_samples_list(self):
        """

        Returns
        -------
        Samples as a list of dictionaries. This is done for compatibility purposes.

        """
        return [vars(sample) for sample in self.samples]

    def generate_timestamp(self):
        """

        Yields
        -------
        A timestamp of a sample, calculated relative to that of the newest sample.

        """
        sampleindex = 0
        while 1:
            yield self.newest_timestamp - sampleindex * self.timeinterval
            sampleindex = sampleindex + 1



import pytest
from wscodec.encoder.pyencoder.instrumented import InstrumentedSampleTRH, InstrumentedSample
from wscodec.decoder import decode
from wscodec.decoder.exceptions import NoCircularBufferError, DelimiterNotFoundError, InvalidMajorVersionError, \
    InvalidFormatError, MessageIntegrityError
from wscodec.decoder.status import SVSH_BIT

INPUT_SERIAL = 'abcdabcd'
INPUT_TIMEINT = 12
INPUT_SECKEY = 'AAAABBBBCCCCDDDD'
INPUT_STATUSB64 = 'MDAwMDAw'
INPUT_CBUFLENBLKS = 32
INPUT_BASEURL = "plotsensor.com"

PADDING = 2


@pytest.fixture(scope="function",
                params=[{'baseurl': "plotsensor.com", 'secretkey': "AAAABBBBCCCCDDDE"},
                        {'baseurl': "toastersrg.plotsensor.com", 'secretkey': "AAAABBBBCCCCDDDF"},
                        {'baseurl': "plotsensor.com", 'secretkey': "AAAABBBBCCCCDDDG"},
                        {'baseurl': "plotsensor.com", 'secretkey': "AAAABBBBCCCCDDDX"}
                        ])
def instr_sample(request):
    return InstrumentedSampleTRH(baseurl=request.param['baseurl'],
                                 serial=INPUT_SERIAL,
                                 secretkey=request.param['secretkey'],
                                 smplintervalmins=INPUT_TIMEINT)


def comparelists(urllist, inlist):
    ONEBIT_ACCURACY = 165 / 4095
    assert len(urllist) == len(inlist)
    for x in range(0, 1, len(inlist)):
        for k, v in inlist[x].items():
            assert inlist[x][k] == pytest.approx(urllist[x][k], abs=ONEBIT_ACCURACY)


def test_lists_equal(instr_sample_populated):
    inlist = instr_sample_populated['inlist']
    urllist = instr_sample_populated['urllist']

    comparelists(urllist, inlist)


@pytest.mark.parametrize('n', range(1, 500, 1))
@pytest.mark.parametrize('usehmac', (True, False))
def test_md5(n, usehmac):
    instr_md5 = InstrumentedSampleTRH(baseurl=INPUT_BASEURL,
                                      serial=INPUT_SERIAL,
                                      secretkey=INPUT_SECKEY,
                                      smplintervalmins=INPUT_TIMEINT,
                                      usehmac=usehmac
                                      )

    inlist = instr_md5.pushsamples(n)

    # Decode the URL
    par = instr_md5.eepromba.get_url_parsedqs()
    decodedurl = decode(secretkey=INPUT_SECKEY, statb64=par['x'][0], timeintb64=par['t'][0],
                        circb64=par['q'][0], vfmtb64=par['v'][0], usehmac=usehmac)

    urllist = decodedurl.get_samples_list()
    for d in urllist:
        del d['timestamp']
        del d['rawtemp']
        del d['rawrh']

    inlist = inlist[:len(urllist)]

    comparelists(urllist, inlist)


def test_batteryvoltage():
    testbatv = 8

    instr_batv = InstrumentedSampleTRH(baseurl=INPUT_BASEURL,
                                       serial=INPUT_SERIAL,
                                       secretkey="",
                                       smplintervalmins=INPUT_TIMEINT,
                                       usehmac=False,
                                       batteryadc=testbatv)

    instr_batv.pushsamples(3)

    # Decode the URL
    par = instr_batv.eepromba.get_url_parsedqs()
    decodedurl = decode(secretkey="", statb64=par['x'][0], timeintb64=par['t'][0],
                        circb64=par['q'][0], vfmtb64=par['v'][0], usehmac=False)

    assert decodedurl.status.get_batvoltageraw() == testbatv


def test_error_nocircularbuffer():
    instr = InstrumentedSample(baseurl=INPUT_BASEURL,
                               serial=INPUT_SERIAL,
                               secretkey="",
                               smplintervalmins=INPUT_TIMEINT,
                               usehmac=False,
                               )

    resetcause = SVSH_BIT
    instr.ffimodule.lib.enc_init(resetcause, True, 0)

    par = instr.eepromba.get_url_parsedqs()

    with pytest.raises(NoCircularBufferError) as excinfo:
        # Attempt to decode the parameters
        decodedurl = decode(secretkey="", statb64=par['x'][0], timeintb64=par['t'][0],
                            circb64="", vfmtb64=par['v'][0], usehmac=False)

    assert excinfo.value.status.resetcause['supervisor'] == True


def test_error_delimiternotfound():
    SECRETKEY = "12345678ABCDEFGH"
    instr = InstrumentedSampleTRH(baseurl=INPUT_BASEURL,
                                  serial=INPUT_SERIAL,
                                  secretkey=SECRETKEY,
                                  smplintervalmins=INPUT_TIMEINT,
                                  usehmac=True,
                                  )

    instr.ffimodule.lib.enc_init(0, False, 0)

    par = instr.eepromba.get_url_parsedqs()

    with pytest.raises(DelimiterNotFoundError) as excinfo:
        # Attempt to decode the parameters
        decodedurl = decode(secretkey=SECRETKEY, statb64=par['x'][0], timeintb64=par['t'][0],
                            circb64=par['q'][0], vfmtb64=par['v'][0], usehmac=True)


def test_error_versionmismatch(mocker):
    instr = InstrumentedSample(baseurl=INPUT_BASEURL,
                               serial=INPUT_SERIAL,
                               secretkey="",
                               smplintervalmins=INPUT_TIMEINT,
                               usehmac=False,
                               )

    BAD_VERSION = 65500

    def mock_version():
        return BAD_VERSION

    instr.ffimodule.lib.enc_init(0, False, 0)

    par = instr.eepromba.get_url_parsedqs()

    mocker.patch('wscodec.decoder.decoderfactory._get_decoderversion', mock_version)

    with pytest.raises(InvalidMajorVersionError) as excinfo:
        # Attempt to decode the parameters
        decodedurl = decode(secretkey="", statb64=par['x'][0], timeintb64=par['t'][0],
                            circb64="", vfmtb64=par['v'][0], usehmac=False)

    assert excinfo.value.decoderversion == BAD_VERSION


def test_error_formatmismatch():
    instr = InstrumentedSample(baseurl=INPUT_BASEURL,
                               serial=INPUT_SERIAL,
                               secretkey="",
                               smplintervalmins=INPUT_TIMEINT,
                               usehmac=False,
                               )

    BAD_FORMAT = 254
    instr.ffimodule.lib.nv.format = bytes([BAD_FORMAT])
    instr.ffimodule.lib.enc_init(0, False, 0)

    par = instr.eepromba.get_url_parsedqs()

    with pytest.raises(InvalidFormatError) as excinfo:
        # Attempt to decode the parameters
        decodedurl = decode(secretkey="", statb64=par['x'][0], timeintb64=par['t'][0],
                            circb64=par['q'][0], vfmtb64=par['v'][0], usehmac=False)

    assert excinfo.value.circformat == BAD_FORMAT


def test_error_messageintegrity():
    instr = InstrumentedSampleTRH(baseurl=INPUT_BASEURL,
                                  serial=INPUT_SERIAL,
                                  secretkey="12345678ABCDEFGH",
                                  smplintervalmins=INPUT_TIMEINT,
                                  usehmac=True,
                                  )

    instr.ffimodule.lib.enc_init(0, False, 0)
    instr.pushsamples(100)

    par = instr.eepromba.get_url_parsedqs()

    with pytest.raises(MessageIntegrityError) as excinfo:
        # Attempt to decode the parameters
        decodedurl = decode(secretkey="notthekey", statb64=par['x'][0], timeintb64=par['t'][0],
                            circb64=par['q'][0], vfmtb64=par['v'][0], usehmac=True)

def test_error_messageintegrity_wrongalgorithm():
    SECRETKEY = "12345678ABCDEFGH"
    instr = InstrumentedSampleTRH(baseurl=INPUT_BASEURL,
                                  serial=INPUT_SERIAL,
                                  secretkey=SECRETKEY,
                                  smplintervalmins=INPUT_TIMEINT,
                                  usehmac=True,
                                  )

    instr.ffimodule.lib.enc_init(0, False, 0)
    instr.pushsamples(100)

    par = instr.eepromba.get_url_parsedqs()

    with pytest.raises(MessageIntegrityError) as excinfo:
        # Attempt to decode the parameters
        decodedurl = decode(secretkey=SECRETKEY, statb64=par['x'][0], timeintb64=par['t'][0],
                            circb64=par['q'][0], vfmtb64=par['v'][0], usehmac=False)


from .b64decode import B64Decoder
from .status import Status
from .exceptions import DelimiterNotFoundError, NoCircularBufferError
from struct import unpack


class CircularBufferURL:
    """
    Base class for a cuplcodec URL.

    This includes at least a circular buffer with a long string of base64 encoded sample data and
    a short status field.

    Instantiation decodes the status string first. It contains error information from the microcontroller running
    the encoder.

    Next it locates the :ref:`ENDSTOP_BYTE` in the circular buffer string. Characters to its left are the newest.
    Characters to its right are the oldest. The circular buffer is unwrapped into a string where :ref:`ENDSTOP_BYTE` is
    the last character and the oldest data is in the first.

    The linearised buffer is further divided into two parts:
    The endstop string (including the endstop itself) are at the end. It contains metadata such as the number of
    samples in the payload. This is preceded by the payload string, which contains a list base64-encocded
    environmental sensor readings. These are in chronological order oldest-to-newest reading left-to-right.

    The decoding of the payload string is handled elsewhere.

    Parameters
        ----------
        statb64 : str
            Base64 encoded status string extract from a URL parameter.
        circb64 : str
            A long string containing base64 encoded samples that are organised as a circular buffer.

    """
    ELAPSED_LEN_BYTES = 4   #: Length of the endstop elapsed minutes field in bytes (including the endstop itself).
    ENDSTOP_LEN_BYTES = 16  #: Length of the endstop in bytes.
    ENDSTOP_BYTE = '~'      #: The last character in the endstop and the end of the circular buffer. Must be URL safe.

    def __init__(self, statb64: str, circb64: str = None):
        self.statb64 = statb64
        self.circb64 = circb64

        self._decode_status()
        self._linearise()

        # Separate the buffer endstop from the payload
        self.endstopstr = self.linearbuf[-self.ENDSTOP_LEN_BYTES:]
        self.payloadstr = self.linearbuf[:-self.ENDSTOP_LEN_BYTES]

        self._decode_endstop()

    def _linearise(self):
        """
        Linearise the circular buffer.

        The circular buffer is made linear by concatenating the two parts of the buffer
        either side of the end stop.
        """

        if len(self.circb64) == 0:
            raise NoCircularBufferError(self.status)

        # Split query string at the end of the endstop marker.
        splitend = self.circb64.split(self.ENDSTOP_BYTE)

        if len(splitend) != 2:
            raise DelimiterNotFoundError(self.circb64, self.status)

        circbufstart = splitend[1]
        circbufend = splitend[0]

        self.endmarkerpos = len(circbufend)
        self.linearbuf = circbufstart + circbufend + self.ENDSTOP_BYTE

    def _decode_status(self):
        """
        Instantiate a Status object. This can be over-ridden by a child of this class
        if the Status data needs to change in future.
        """
        self.status = Status(self.statb64)

    def _decode_endstop(self):
        """
        Decode the circular buffer endstop. This can be over-ridden by a child of this class
        if the endstop data needs to change in future.
        """
        endstopstr = self.endstopstr

        assert len(endstopstr) == self.ENDSTOP_LEN_BYTES

        endstopstr = endstopstr.replace(self.ENDSTOP_BYTE, B64Decoder.RFC3548_PADDING_BYTE)

        # Extract elapsed minutes since the previous sample in minutes xxx~. Replace the '=' to make this valid base64.
        elapsedb64 = endstopstr[-self.ELAPSED_LEN_BYTES:]
        # The remaining 12 bytes contain the MD5 hash and the number of valid pairs
        hashnb64 = endstopstr[:-self.ELAPSED_LEN_BYTES]

        # Decode elapsedb64
        elapsedbytes = B64Decoder.b64decode(elapsedb64)
        # Decode hashnb64
        hashn = B64Decoder.b64decode(hashnb64)

        # Extract the number of samples and the HMAC/MD5 checksum from the endstop.
        npairsbytes = hashn[7:9]
        hashbytes = hashn[0:7]

        self.elapsedmins = int.from_bytes(elapsedbytes, byteorder='little')
        self.npairs = unpack(">H", npairsbytes)[0]
        self.hash = hashbytes.hex()





from .sharedobj import ndefpy, demipy, pairhistpy, samplepy
from . import eeprom as eeprom
from struct import pack


class InstrumentedBase(object):
    """
    Put some documentation here
    """

    def __init__(self,
                 ffimodule,
                 baseurl,
                 serial,
                 secretkey,
                 smplintervalmins,
                 batteryadc=100,
                 format=1,
                 resetsalltime=0,
                 usehmac=True,
                 httpsdisable=False
                 ):
        # To remove ambiguity, boolean variables are stored as integers.
        if httpsdisable:
            httpsdisable_int = 1
        else:
            httpsdisable_int = 0

        if usehmac:
            usehmac_int = 1
        else:
            usehmac_int = 0

        self.secretkey = secretkey
        self.ffimodule = ffimodule
        self.ffimodule.lib.nv.serial = serial.encode('ascii')
        self.ffimodule.lib.nv.seckey = secretkey.encode('ascii')
        self.ffimodule.lib.nv.baseurl = baseurl.encode('ascii')
        self.ffimodule.lib.nv.format = bytes([format])
        self.ffimodule.lib.nv.usehmac = usehmac_int
        self.ffimodule.lib.nv.httpsdisable = httpsdisable_int
        self.ffimodule.lib.nv.resetsalltime = resetsalltime
        smplintbytes = pack("<H", smplintervalmins)
        self.ffimodule.lib.nv.smplintervalmins = smplintbytes
        self.eepromba = eeprom.Eeprom(64)


        @self.ffimodule.ffi.def_extern()
        def nt3h_writetag(eepromblk, blkdata):
            self.eepromba.write_block(eepromblk, blkdata)
            return 0

        @self.ffimodule.ffi.def_extern()
        def nt3h_readtag(eepromblk, blkdata):
            self.eepromba.read_block(eepromblk, blkdata)
            return 0

        @self.ffimodule.ffi.def_extern()
        def nt3h_eepromwritedone():
            return 0

        @self.ffimodule.ffi.def_extern()
        def printint(myint):
            print("printint " + str(myint))
            return 0



class InstrumentedNDEF(InstrumentedBase):
    def __init__(self,
                 baseurl='plotsensor.com',
                 serial='AAAACCCC',
                 secretkey='AAAACCCC',
                 smplintervalmins=12):
        super(InstrumentedNDEF, self).__init__(ndefpy, baseurl, serial, secretkey, smplintervalmins)


class InstrumentedDemi(InstrumentedBase):
    def __init__(self,
                 baseurl='plotsensor.com',
                 serial='AAAACCCC',
                 secretkey='AAAACCCC',
                 smplintervalmins=12):
        super(InstrumentedDemi, self).__init__(demipy, baseurl, serial, secretkey, smplintervalmins)


class InstrumentedPairhist(InstrumentedBase):
    def __init__(self,
                 baseurl='plotsensor.com',
                 serial='AAAACCCC',
                 secretkey='AAAACCCC',
                 smplintervalmins=12):
        super(InstrumentedPairhist, self).__init__(pairhistpy, baseurl, serial, secretkey, smplintervalmins)


class InstrumentedSample(InstrumentedBase):
    def __init__(self,
                 baseurl='plotsensor.com',
                 serial='AAAACCCC',
                 secretkey='AAAACCCC',
                 smplintervalmins=12,
                 batteryadc=100,
                 format=1,
                 resetsalltime=0,
                 usehmac=True,
                 httpsdisable=False
                 ):
        super(InstrumentedSample, self).__init__(samplepy, baseurl, serial, secretkey, smplintervalmins, batteryadc, format, resetsalltime, usehmac, httpsdisable)
        self.set_battery_adc(batteryadc)

        @self.ffimodule.ffi.def_extern()
        def batv_measure():
            return self.batteryadc

    def set_battery_adc(self, batteryadc):
        self.batteryadc = batteryadc

    def get_url(self):
        return self.eepromba.get_url()

    def temp_degc_to_raw(self, degc):
        """ Converts degrees C to a raw ADC value for the Texas HDC2010. """
        return int((degc + 40) * 4096 / 165)


    def rh_percent_to_raw(self, rhpc):
        """ Converts from relative humidity in percent to a raw ADC value for the Texas HDC2010. """
        return int((rhpc * 4096) / 100)

    def tempsample(self, countermax, counterstep):
        counter = 0
        while True:
            counter += counterstep
            counter = counter % countermax
            rawtemp = self.temp_degc_to_raw(counter)
            yield {'adc': rawtemp, 'ref': counter}

    def rhsample(self, countermax, counterstep):
        counter = 0
        while True:
            counter += counterstep
            counter = counter % countermax
            rawrh = self.rh_percent_to_raw(counter)
            yield {'adc': rawrh, 'ref': counter}


class InstrumentedSampleT(InstrumentedSample):
    def __init__(self,
                 serial='ABCDEFGH',
                 secretkey='AAAACCCC11112222',
                 baseurl='plotsensor.com',
                 smplintervalmins=12,
                 resetsalltime=0,
                 batteryadc=100,
                 resetcause=0,
                 usehmac=True,
                 httpsdisable=False,
                 tagerror=False
                 ):
        super(InstrumentedSampleT, self).__init__(baseurl,
                                                  serial,
                                                  secretkey,
                                                  smplintervalmins,
                                                  batteryadc=batteryadc,
                                                  format=2,
                                                  resetsalltime=resetsalltime,
                                                  usehmac=usehmac,
                                                  httpsdisable=httpsdisable)
        self.ffimodule.lib.enc_init(resetcause, tagerror, 0)

    def pushsamples(self, num):
        inlist = list()
        tempgen = self.tempsample(100, 0.3)
        for i in range(0, num):
            tempsmpl = next(tempgen)
            inlist.insert(0, {'temp': tempsmpl['ref']})
            self.ffimodule.lib.enc_pushsample(tempsmpl['adc'], 0)
        return inlist


class InstrumentedSampleTRH(InstrumentedSample):
    def __init__(self,
                 serial='ABCDEFGH',
                 secretkey='AAAACCCC11112222',
                 baseurl='plotsensor.com',
                 smplintervalmins=12,
                 resetsalltime=0,
                 batteryadc=100,
                 resetcause=0,
                 usehmac=True,
                 httpsdisable=False,
                 tagerror=False
                 ):
        super(InstrumentedSampleTRH, self).__init__(baseurl,
                                                    serial,
                                                    secretkey,
                                                    smplintervalmins,
                                                    batteryadc=batteryadc,
                                                    format=1,
                                                    resetsalltime=resetsalltime,
                                                    usehmac=usehmac,
                                                    httpsdisable=httpsdisable)
        self.ffimodule.lib.enc_init(resetcause, tagerror, 0)

    def pushsamples(self, num):
        inlist = list()
        tempgen = self.tempsample(100, 1.0)
        rhgen = self.rhsample(100, 1.0)
        for i in range(0, num):
            tempsmpl = next(tempgen)
            rhsmpl = next(rhgen)
            inlist.insert(0, {'temp': tempsmpl['ref'], 'rh': rhsmpl['ref']})
            self.ffimodule.lib.enc_pushsample(tempsmpl['adc'], rhsmpl['adc'])
        return inlist

    def pushsamplelist(self, trhlist: list):
        """

        :param trhlist: a list of dictionaries each containing temperature and relative humidity keys.
        :return: None
        """
        for smpldict in trhlist:
            tempdegc = smpldict['temp']
            rhpc = smpldict['rh']
            tempraw = self.temp_degc_to_raw(tempdegc)
            rhraw = self.rh_percent_to_raw(rhpc)
            self.ffimodule.lib.enc_pushsample(tempraw, rhraw)

    def updateendstop(self, minutes: int):
        """ Update the endstop with minutes elapsed since the most recent sample.

        :param minutes: Minutes elapsed since the most recent sample.
        :return: None
        """
        self.ffimodule.lib.enc_setelapsed(minutes)

    def geturlqs(self):
        return self.eepromba.get_url_parsedqs()



import pytest
from wscodec.encoder.pyencoder.instrumented import InstrumentedPairhist
import random
import hmac
import hashlib

INPUT_SERIAL = 'abcdabcd'
INPUT_TIMEINT = 12
INPUT_SECKEY = 'AAAABBBBCCCCDDDD'
INPUT_STATUSB64 = 'MDAwMDAw'
INPUT_CBUFLENBLKS = 32

PADDING = 2
# This is for loopcount, resetsalltime, batv_resetcause and endmarkerpos.
# Not part of the test so are set to zero.
LOOPCOUNT = 0
RESETSALLTIME = 0
BATV_RESETCAUSE = 0
ENDMARKERPOS = 0
ZERO_STAT_AND_ENDMARKERPOS = bytearray(8)


@pytest.fixture(scope="function",
                params=[{'baseurl': "plotsensor.com", 'secretkey': "AAAACCCCDDDDEEEE"},
                        {'baseurl': "toastersrg.plotsensor.com", 'secretkey': "AAAABBBBCCCCDEDE"},
                        {'baseurl': "plotsensor.com", 'secretkey': "masf349212345678"},
                        {'baseurl': "plotsensor.com", 'secretkey': "42r3223355778899"}
                        ])
def instr_pairhist(request):
    return InstrumentedPairhist(baseurl=request.param['baseurl'],
                                serial=INPUT_SERIAL,
                                secretkey=request.param['secretkey'],
                                smplintervalmins=INPUT_TIMEINT)

def convertToSdChars(rd0, rd1):
    rd0Msb = (rd0 >> 4)
    rd1Msb = (rd1 >> 4)
    Lsb = (((rd0 & 0xF) << 4) & (rd1 & 0xF))
    return rd0Msb, rd1Msb, Lsb


def write_buffer(instr, n):
    sdlist = list()
    sdbytearray = bytearray()
    for i in range(0, n):
        testsample = random.randrange(4095)
        rd0Msb, rd1Msb, Lsb = convertToSdChars(testsample, testsample)
        sensordata = instr.ffimodule.ffi.new("pair_t *", {'rd0Msb': rd0Msb, 'rd1Msb': rd1Msb, 'Lsb': Lsb})
        sdbytearray = bytearray([rd0Msb, rd1Msb, Lsb]) + sdbytearray
        sdlist.insert(0, sensordata[0])
        instr.ffimodule.lib.pairhist_push(sensordata[0])
    return sdlist, sdbytearray


def check_buffer(instr, sdlist):
    bufindex = 0
    errorflag = instr.ffimodule.ffi.new("int *", 0)
    for sensordata in sdlist:
        bufsd = instr.ffimodule.lib.pairhist_read(bufindex, errorflag)
        bufindex = bufindex + 1
        assert bufsd.rd0Msb == sensordata.rd0Msb and bufsd.rd1Msb == sensordata.rd1Msb and bufsd.Lsb == sensordata.Lsb


@pytest.fixture(scope="function", params=[1, 10, 100, -1])
def instr_pairhist_with_samples(instr_pairhist, request):
    if request.param == -1:
        nsamples = instr_pairhist.ffimodule.lib.buflenpairs
    else:
        nsamples = request.param
    sdlist, sdba = write_buffer(instr_pairhist, n=nsamples)
    return {'instr_pairhist': instr_pairhist, 'sdlist': sdlist, 'sdba': sdba, 'nsamples': nsamples}


def get_uut_digest(instr_pairhist_with_samples, hmac=False):
    instr_pairhist = instr_pairhist_with_samples['instr_pairhist']
    nsamples = instr_pairhist_with_samples['nsamples']

    uutdigest = instr_pairhist.ffimodule.lib.pairhist_hash(nsamples,
                                                          hmac,
                                                          LOOPCOUNT,
                                                          RESETSALLTIME,
                                                          BATV_RESETCAUSE,
                                                          ENDMARKERPOS).hash
    lendigest = len(uutdigest)

    uutdigeststr = ""
    for i in range(0, lendigest):
        # https://stackoverflow.com/questions/12638408/decorating-hex-function-to-pad-zeros
        uutdigeststr += f"{uutdigest[i]:0{PADDING}x}"

    return uutdigeststr


def test_buffer_load(instr_pairhist_with_samples):
    check_buffer(instr_pairhist_with_samples['instr_pairhist'], instr_pairhist_with_samples['sdlist'])


def test_hmac(instr_pairhist_with_samples):
    secretkey = instr_pairhist_with_samples['instr_pairhist'].secretkey
    secretkeyba = bytearray(secretkey.encode('utf-8'))

    uutdigest = get_uut_digest(instr_pairhist_with_samples, hmac=True)
    sdba = instr_pairhist_with_samples['sdba'] + ZERO_STAT_AND_ENDMARKERPOS

    hmacobj = hmac.new(secretkeyba, sdba, 'md5')
    tbdigest = hmacobj.hexdigest()
    tbdigest = tbdigest[0:len(uutdigest)]

    assert tbdigest == uutdigest


def test_hmac_fail(instr_pairhist_with_samples):
    secretkey = bytes(b"12345678")

    uutdigest = get_uut_digest(instr_pairhist_with_samples, hmac=True)
    sdba = instr_pairhist_with_samples['sdba']

    hmacobj = hmac.new(secretkey, sdba, 'md5')
    tbdigest = hmacobj.hexdigest()
    tbdigest = tbdigest[0:len(uutdigest)]

    assert tbdigest != uutdigest


def test_md5(instr_pairhist_with_samples):
    uutdigest = get_uut_digest(instr_pairhist_with_samples, hmac=False)
    sdba = instr_pairhist_with_samples['sdba']

    m = hashlib.md5()
    m.update(sdba)
    m.update(ZERO_STAT_AND_ENDMARKERPOS) # PAD WITH 8 ZEROES
    tbdigest = m.hexdigest()
    tbdigest = tbdigest[0:len(uutdigest)]

    assert tbdigest == uutdigest






import pytest
from wscodec.encoder.pyencoder.instrumented import InstrumentedSampleT
from wscodec.decoder import decode

INPUT_BASEURL = "cupl.uk"
INPUT_SERIAL = 'abcdabcd'
INPUT_TIMEINT = 12
INPUT_SECKEY = 'AAAABBBBCCCCDDDD'
INPUT_STATUSB64 = 'MDAwMDAw'
INPUT_CBUFLENBLKS = 32

PADDING = 2


@pytest.fixture(scope="function",
                params=[{'baseurl': "plotsensor.com", 'secretkey': INPUT_SECKEY}
                        ])
def instr_sample(request):
    return InstrumentedSampleT(baseurl=request.param['baseurl'],
                               serial=INPUT_SERIAL,
                               secretkey=request.param['secretkey'],
                               smplintervalmins=INPUT_TIMEINT)


def comparelists(urllist, inlist):
    ONEBIT_ACCURACY = 165 / 4096
    assert len(urllist) == len(inlist)
    for x in range(0, 1, len(inlist)):
        for k, v in inlist[x].items():
            assert inlist[x][k] == pytest.approx(urllist[x][k], abs=ONEBIT_ACCURACY)


def test_lists_equal(instr_sample_populated):
    inlist = instr_sample_populated['inlist']
    urllist = instr_sample_populated['urllist']

    comparelists(urllist, inlist)


@pytest.mark.parametrize('n', range(1, 500, 1))
def test_md5(instr_sample, n):
    instr_md5 = InstrumentedSampleT(baseurl=INPUT_BASEURL,
                                    serial=INPUT_SERIAL,
                                    secretkey="",
                                    smplintervalmins=INPUT_TIMEINT,
                                    usehmac=False
                                    )

    inlist = instr_md5.pushsamples(n)

    # Decode the URL
    par = instr_md5.eepromba.get_url_parsedqs()
    decodedurl = decode(secretkey="", statb64=par['x'][0], timeintb64=par['t'][0],
                        circb64=par['q'][0], vfmtb64=par['v'][0], usehmac=False)

    urllist = decodedurl.get_samples_list()
    for d in urllist:
        del d['timestamp']
        del d['rawtemp']

    inlist = inlist[:len(urllist)]

    comparelists(urllist, inlist)



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cuplcodec",
    version="2.0.3",
    author="Malcolm Mackay",
    author_email="malcolm@plotsensor.com",
    description="Package for creating and decoding URLs that contain temperature and humidity samples.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cuplsensor/cuplcodec",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "License :: Other/Proprietary License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    setup_requires=["cffi>=1.0.0"],
    cffi_modules=["wscodec/encoder/pyencoder/sample_builder.py:ffibuilder",
                  "wscodec/encoder/pyencoder/ndef_builder.py:ffibuilder",
                  "wscodec/encoder/pyencoder/demi_builder.py:ffibuilder",
                  "wscodec/encoder/pyencoder/pairhist_builder.py:ffibuilder"],
    install_requires=["cffi>=1.0.0", "ndeflib>=0.3.2"],
)



import pytest
from base64 import urlsafe_b64decode
from wscodec.encoder.pyencoder.instrumented import InstrumentedNDEF
from wscodec.decoder.exceptions import DelimiterNotFoundError
from wscodec.decoder import decode
from urllib.parse import urlparse

INPUT_SERIAL = 'abcdabcd'
INPUT_TIMEINT = 12
INPUT_SECKEY = 'AAAABBBBCCCCDDDD'
INPUT_STATUSB64 = 'MDAwMDAw'
INPUT_CBUFLENBLKS = 32


@pytest.fixture(scope="function",
                params=["plotsensor.com",
                        "toastersrg.plotsensor.com",
                        "0202u390ur2309323232223233232u20392039209302.plotsensor.com"])
def instr_ndef(request):
    ndefobj = InstrumentedNDEF(baseurl=request.param, serial=INPUT_SERIAL, secretkey=INPUT_SECKEY,
                               smplintervalmins=INPUT_TIMEINT)
    ndefobj.baseurl = request.param
    return ndefobj


@pytest.fixture
def makeblankurl(instr_ndef):
    ndefobj = instr_ndef
    cbuflenblks = INPUT_CBUFLENBLKS
    statusb64bytes = INPUT_STATUSB64.encode('ascii')
    statusb64chars = ndefobj.ffimodule.ffi.new("char[]", statusb64bytes)
    startblkptr = ndefobj.ffimodule.ffi.new("int *", 0)
    ndefobj.ffimodule.lib.ndef_writeblankurl(cbuflenblks, statusb64chars, startblkptr)
    return ndefobj


@pytest.fixture
def blankurlqs(makeblankurl):
    # Obtain the URL parameters dictionary from the Mock EEPROM
    return makeblankurl.eepromba.get_url_parsedqs()


@pytest.fixture
def blankurlraw(makeblankurl):
    # Obtain the URL parameters dictionary from the Mock EEPROM
    return makeblankurl.eepromba.get_message()


def test_calclen(instr_ndef):
    ndefobj = instr_ndef
    paddinglen = ndefobj.ffimodule.ffi.new("int *")
    preamblenbytes = ndefobj.ffimodule.ffi.new("int *")
    urllen = ndefobj.ffimodule.ffi.new("int *")
    paddinglen[0] = 0
    preamblenbytes[0] = 0
    urllen[0] = len(ndefobj.baseurl)
    ndefobj.ffimodule.lib.ndef_calclen(paddinglen, preamblenbytes, urllen)
    x = preamblenbytes[0]
    assert (x) % 16 == 0


def test_baseurl(makeblankurl):
    blankurl = makeblankurl.eepromba.get_url()
    baseurl = urlparse(blankurl).netloc
    assert baseurl == makeblankurl.baseurl


def test_serial(blankurlqs):
    serial = blankurlqs['s'][0]
    assert serial == INPUT_SERIAL


def test_timeinterval(blankurlqs):
    timeintb64 = blankurlqs['t'][0]
    timeintb64 = timeintb64.replace('.', '=')
    timeintbytes = urlsafe_b64decode(timeintb64)
    timeint = int.from_bytes(timeintbytes, byteorder='little')
    assert timeint == INPUT_TIMEINT


def test_statusbytes(blankurlqs):
    statb64 = blankurlqs['x'][0]
    assert statb64 == INPUT_STATUSB64


def test_circbuf_length(blankurlqs):
    circb64 = blankurlqs['q'][0]
    blksizebytes = 16
    assert len(circb64) / blksizebytes == INPUT_CBUFLENBLKS


def test_decode_raises_indexerrror(blankurlqs):
    timeintb64 = blankurlqs['t'][0]
    statb64 = blankurlqs['x'][0]
    circb64 = blankurlqs['q'][0]
    ver = blankurlqs['v'][0]
    with pytest.raises(DelimiterNotFoundError):
        # Attempt to decode the parameters
        decoded = decode(secretkey=INPUT_SECKEY,
                         timeintb64=timeintb64,
                         statb64=statb64,
                         circb64=circb64,
                         vfmtb64=ver)



import base64


class B64Decoder:
    URLSAFE_PADDING_BYTE = '.'
    RFC3548_PADDING_BYTE = '='

    @classmethod
    def b64decode(cls, b64string):
        # Replace padding byte with RFC3548
        b64string = b64string.replace(cls.URLSAFE_PADDING_BYTE, cls.RFC3548_PADDING_BYTE)
        return base64.urlsafe_b64decode(b64string)


import pytest
from wscodec.encoder.pyencoder.instrumented import InstrumentedSampleTRH
from wscodec.decoder import decode

INPUT_SERIAL = 'abcdabcd'
INPUT_TIMEINT = 12
INPUT_SECKEY = 'AAAABBBBCCCCDDDD'
INPUT_STATUSB64 = 'MDAwMDAw'
INPUT_CBUFLENBLKS = 32
INPUT_BASEURL = "plotsensor.com"

PADDING = 2


@pytest.fixture(scope="function",
                params=[{'baseurl': INPUT_BASEURL, 'secretkey': INPUT_SECKEY}
                        ])
def instr_sample(request):
    return InstrumentedSampleTRH(baseurl=request.param['baseurl'],
                                 serial=INPUT_SERIAL,
                                 secretkey=request.param['secretkey'],
                                 smplintervalmins=INPUT_TIMEINT)


def test_status(instr_sample, n=1):
    instr_sample.pushsamples(n)
    encoder_cursorpos = instr_sample.ffimodule.lib.demi_getendmarkerpos()

    # Decode the URL
    par = instr_sample.eepromba.get_url_parsedqs()
    decodedurl = decode(secretkey=instr_sample.secretkey, statb64=par['x'][0], timeintb64=par['t'][0],
                        circb64=par['q'][0], vfmtb64=par['v'][0])

    assert encoder_cursorpos == decodedurl.endmarkerpos


@pytest.mark.parametrize('n', [1, 300])
def test_loopcount(instr_sample, n):
    instr_sample.pushsamples((instr_sample.ffimodule.lib.buflenpairs + 4) * n + 1)

    par = instr_sample.eepromba.get_url_parsedqs()
    decodedurl = decode(secretkey=instr_sample.secretkey, statb64=par['x'][0], timeintb64=par['t'][0],
                        circb64=par['q'][0], vfmtb64=par['v'][0])

    assert decodedurl.status.loopcount == n


@pytest.mark.parametrize('resetsalltime', [1, 300])
def test_resetsalltime(resetsalltime):
    instr_sample = InstrumentedSampleTRH(baseurl=INPUT_BASEURL,
                                         serial=INPUT_SERIAL,
                                         secretkey=INPUT_SECKEY,
                                         smplintervalmins=INPUT_TIMEINT,
                                         resetsalltime=resetsalltime)

    instr_sample.pushsamples(1)

    par = instr_sample.eepromba.get_url_parsedqs()
    decodedurl = decode(secretkey=instr_sample.secretkey, statb64=par['x'][0], timeintb64=par['t'][0],
                        circb64=par['q'][0], vfmtb64=par['v'][0])

    assert decodedurl.status.resetsalltime == resetsalltime


@pytest.mark.parametrize('resetcause', [1, 100, 254])
@pytest.mark.parametrize('batteryadc', [1, 100, 254])
def test_batteryvoltage(resetcause, batteryadc):
    instr_sample = InstrumentedSampleTRH(batteryadc=batteryadc)
    # Hack. Sorry. The def_extern must be defined beforehand but this is not possible if instr_sample does not exist.
    # Any better way of doing this please PR.

    instr_sample = InstrumentedSampleTRH(baseurl=INPUT_BASEURL,
                                         serial=INPUT_SERIAL,
                                         secretkey=INPUT_SECKEY,
                                         smplintervalmins=INPUT_TIMEINT,
                                         batteryadc=batteryadc,
                                         resetcause=resetcause)

    instr_sample.pushsamples(2)

    par = instr_sample.eepromba.get_url_parsedqs()
    decodedurl = decode(secretkey=instr_sample.secretkey, statb64=par['x'][0], timeintb64=par['t'][0],
                        circb64=par['q'][0], vfmtb64=par['v'][0])

    assert decodedurl.status.get_batvoltageraw() == batteryadc
    assert decodedurl.status.get_resetcauseraw() == resetcause
